require 'random_graph/random_generators'
require 'random_graph/graph_algorithms'
require 'graphviz'

module RandomGraph
  # Graph class implements basic graph object and relevant instance methods
  class Graph
    include Enumerable
    extend RandomGenerators
    include GraphAlgorithms

    attr_reader :size
    attr_accessor :nodes
    attr_accessor :viz_counter

    # We will used the adjacency list representation of a graph
    # default to size 0
    def initialize(size = 0)
      @viz_counter = 0
      @size = size
      @nodes = Array.new(size) { [] }
    end

    def visualize
      g = GraphViz.new(:G, type: :graph)
      real_nodes = Array.new(@size)
      (0...@size).each { |i| real_nodes[i] = g.add_nodes("#{i}") }
      (0...@size).each do |i|
        @nodes[i].each do |n|
          g.add_edges(real_nodes[i], real_nodes[n]) if i < n
        end
      end
      g.output(png: "graph_#{@viz_counter}.png")
      @viz_counter += 1
    end

    def add_node
      new_node = []
      @nodes << new_node
      @size += 1
    end

    def each
      @nodes.each { |v| yield v }
    end

    # add edge between two nodes
    def add_edge(from, to)
      fail ArgumentError, 'invalid node' if from >= @size || from < 0
      fail ArgumentError, 'invalid node' if to >= @size || to < 0
      return if to == from
      @nodes[from] << to unless @nodes[from].include?(to)
      @nodes[to] << from unless @nodes[to].include?(from)
    end

    def remove_edge(from, to)
      @nodes[from].delete(to) if @nodes[from].include?(to)
      @nodes[to].delete(from) if @nodes[to].include?(from)
    end

    def degree(node)
      @nodes[node].size
    end

    def max_degree
      max_deg = 0
      @nodes.each do |v|
        max_deg = v.size if max_deg < v.size
      end
      max_deg
    end

    def adjacent_nodes(v)
      @nodes[v]
    end

    # is there an edge between two nodes?
    def edge?(from, to)
      # check if nodes are valid?
      invalid_nodes = from >= @size || from < 0 || to >= @size || to < 0
      @nodes[from].include?(to) unless invalid_nodes
    end

    def edge_number
      count = 0
      @nodes.each do |n|
        count += n.size
      end
      count / 2
    end

    def average_degree
      2.0 * edge_number / size
    end
  end
end
