module RandomGraph
  # implements random graph generators
  module RandomGenerators
    # Select a graph randomly, where all graphs with n nodes and m edges
    # are equally likely to be selected.
    # O(n^2) expected worst-case rumtime, assuming random numbers generated
    # in constant time. Number of random edges probed before adding edge k
    # is (n^2 - n)/(n^2 - n - 2k). Total expected number of probes can be
    # express as a summation, which is upper bounded by O(n^2)

    def erdos_renyi_gnm(n, m)
      fail ArgumentError, 'must have positive number of nodes' if n < 0
      fail ArgumentError, 'must have positive number of edges' if m < 0

      possible_edges = n * (n - 1) / 2

      fail ArgumentError, 'too many edges' if m > possible_edges

      # create empty graph
      g = Graph.new(n)

      # for every edge
      (0...m).each do |_i|
        loop do
          # select a random number, representing on of the possible edges
          from = rand(0...n)
          to = rand(0...n)
          break if !g.edge?(from, to) && (from != to)
        end
        g.add_edge(from, to)
      end
      g
    end

    # Create a graph with n nodes, where all edges have a probability p
    # of existing
    def erdos_renyi_gnp(n, p)
      fail ArgumentError, 'must have positive number of nodes' if n < 0
      fail ArgumentError, 'need positive probability' if p < 0
      fail ArgumentError, 'probability cannot be greater than 1!' if p > 1

      # create empty graph
      g = Graph.new(n)
      (0...n**2).each do |i|
        from = i / n
        to = i % n
        g.add_edge(from, to) if rand <= p && to > from
      end
      g
    end

    # creates a watts-strogatz random graph. n represents the number of nodes,
    # k is the mean degree, and b is the beta parameter
    def watts_strogatz(n, k, b)
      fail ArgumentError, 'must have positive number of nodes' if n < 0
      fail ArgumentError, 'invalid beta parameter' if b < 0 || b > 1
      fail ArgumentError, 'invalid mean degree if' if k > n || k < Math.log10(n)

      # construction of regular ring lattice
      g = Graph.new(n)
      (0...n).each do |i|
        (0...n).each do |j|
          ab = (i - j).abs
          modded = n - 1 - (k / 2)
          checker = ab % modded
          g.add_edge(i, j) if 0 < checker && checker <= (k / 2)
        end
      end

      # rewiring under beta parameter
      (0...n).each do |i|
        (i...n).each do |j|
          # skip non-existent edges
          next unless g.edge?(i, j)
          # rewire from uniform distribution
          if rand < b && g.degree(i) != n - 1
            # skip is node is completely wired
            target = 0
            loop do
              target = rand(0...n)
              break if target != i && !g.edge?(i, target)
            end
            g.remove_edge(i, j)
            g.add_edge(i, target)
          end
        end
      end
      g
    end

    # preferential attachement model enforces power law degree distribution
    def preferential_attachment(size)
      fail ArgumentError, 'invalid number of nodes' if size < 2
      g = Graph.new(2)
      g.add_edge(0, 1)
      total_degree = 2

      (2...size).each do |i|
        g.add_node
        r = rand(1..total_degree)
        j = 0

        loop do
          r -= g.degree(j)
          break if r <= 0
          j += 1
        end
        g.add_edge(i, j)
        total_degree += 2
      end
      g
    end

    # n x n grid, with vertices isolated with probability p

    def random_grid(n, p)
      fail ArgumentError, 'invalid probability' if p > 1 || p < 0
      g = Graph.new(n**2)
      (0...n**2).each do |i|
        bottom = (i + n) % n**2
        right = (i + 1)
        right -= n if right % n == 0
        g.add_edge(i, bottom)
        g.add_edge(i, right)
      end
      (0...n**2).each do |i|
        next if rand > p
        top = i - n
        top += n**2 if top < 0
        bottom = (i + n) % n**2
        right = i + 1
        right -= n if right % n == 0
        left = i - 1
        left += n if left == -1 || left % n == n - 1
        g.remove_edge(i, top)
        g.remove_edge(i, bottom)
        g.remove_edge(i, left)
        g.remove_edge(i, right)
      end
      g
    end
  end
end
